/*
 * kgraft_patch_bsc1189278
 *
 * Fix for CVE-2021-38198, bsc#1189278
 *
 *  Upstream commit:
 *  b1bd5cba3306 ("KVM: X86: MMU: Use the correct inherited permissions to get
 *                 shadow page")
 *
 *  SLE12-SP3 commit:
 *  07db41a66b3d7c9f8d3fc3487c43d349fcd77a1b
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  41b872da6e79a89c0544a9a6fe7175e86f06bdd4
 *
 *  SLE15-SP2 and -SP3 commit:
 *  1a0a225f892118e5ca56e5d299f9e5cdb685acc6
 *
 *
 *  Copyright (c) 2021 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64)

#if !IS_MODULE(CONFIG_KVM)
#error "Live patch supports only CONFIG_KVM=m"
#endif

#include <linux/kernel.h>
#include <linux/tracepoint.h>

/* from include/linux/tracepoint.h */
#define KGRR___DECLARE_TRACE(name, proto, args, cond, data_proto, data_args) \
	static struct tracepoint (*kgre___tracepoint_##name);		\
	static inline void kgrr_trace_##name(proto)			\
	{								\
	if (unlikely(static_key_enabled(&(*kgre___tracepoint_##name).key))) \
		__DO_TRACE(&(*kgre___tracepoint_##name),		\
				TP_PROTO(data_proto),			\
				TP_ARGS(data_args),			\
				TP_CONDITION(cond),,);			\
		if (IS_ENABLED(CONFIG_LOCKDEP) && (cond)) {		\
			rcu_read_lock_sched_notrace();			\
			rcu_dereference_sched((*kgre___tracepoint_##name).funcs); \
			rcu_read_unlock_sched_notrace();		\
		}							\
	}

#define KGRR_DECLARE_TRACE(name, proto, args)				\
	KGRR___DECLARE_TRACE(name, PARAMS(proto), PARAMS(args),		\
			cpu_online(raw_smp_processor_id()),		\
			PARAMS(void *__data, proto),			\
			PARAMS(__data, args))

#define KGRR_TRACE_EVENT(name, proto, args)			\
	KGRR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))



/* klp-ccp: from arch/x86/kvm/irq.h */
#include <linux/mm_types.h>
#include <linux/hrtimer.h>
#include <linux/kvm_host.h>

/* klp-ccp: from arch/x86/include/asm/kvm_host.h */
static struct kvm_x86_ops *(*kgre_kvm_x86_ops);

/* klp-ccp: from include/linux/kvm_host.h */
static void (*kgre_kvm_release_page_dirty)(struct page *page);

static void (*kgre_kvm_release_pfn_clean)(kvm_pfn_t pfn);

static unsigned long (*kgre_kvm_vcpu_gfn_to_hva_prot)(struct kvm_vcpu *vcpu, gfn_t gfn, bool *writable);

static int (*kgre_kvm_vcpu_read_guest_atomic)(struct kvm_vcpu *vcpu, gpa_t gpa, void *data,
			       unsigned long len);

static void (*kgre_kvm_vcpu_mark_page_dirty)(struct kvm_vcpu *vcpu, gfn_t gfn);

static void (*kgre_kvm_flush_remote_tlbs)(struct kvm *kvm);

/* klp-ccp: from arch/x86/kvm/irq.h */
#include <linux/spinlock.h>

/* klp-ccp: from arch/x86/kvm/ioapic.h */
#include <linux/kvm_host.h>
#include <kvm/iodev.h>

#define ASSERT(x) do { } while (0)

/* klp-ccp: from arch/x86/kvm/lapic.h */
#include <kvm/iodev.h>
#include <linux/kvm_host.h>
/* klp-ccp: from arch/x86/kvm/mmu.h */
#include <linux/kvm_host.h>

/* klp-ccp: from arch/x86/kvm/kvm_cache_regs.h */
#define KVM_POSSIBLE_CR0_GUEST_BITS X86_CR0_TS
#define KVM_POSSIBLE_CR4_GUEST_BITS				  \
	(X86_CR4_PVI | X86_CR4_DE | X86_CR4_PCE | X86_CR4_OSFXSR  \
	 | X86_CR4_OSXMMEXCPT | X86_CR4_PGE)

static inline ulong kgrr_kvm_read_cr0_bits(struct kvm_vcpu *vcpu, ulong mask)
{
	ulong tmask = mask & KVM_POSSIBLE_CR0_GUEST_BITS;
	if (tmask & vcpu->arch.cr0_guest_owned_bits)
		(*kgre_kvm_x86_ops)->decache_cr0_guest_bits(vcpu);
	return vcpu->arch.cr0 & mask;
}

static inline ulong kgrr_kvm_read_cr4_bits(struct kvm_vcpu *vcpu, ulong mask)
{
	ulong tmask = mask & KVM_POSSIBLE_CR4_GUEST_BITS;
	if (tmask & vcpu->arch.cr4_guest_owned_bits)
		(*kgre_kvm_x86_ops)->decache_cr4_guest_bits(vcpu);
	return vcpu->arch.cr4 & mask;
}

static inline u32 kgrr_kvm_read_pkru(struct kvm_vcpu *vcpu)
{
	return (*kgre_kvm_x86_ops)->get_pkru(vcpu);
}

/* klp-ccp: from arch/x86/kvm/mmu.h */
#define PT64_PT_BITS 9
#define PT64_ENT_PER_PAGE (1 << PT64_PT_BITS)

#define PT_WRITABLE_SHIFT 1
#define PT_USER_SHIFT 2

#define PT_PRESENT_MASK (1ULL << 0)
#define PT_WRITABLE_MASK (1ULL << PT_WRITABLE_SHIFT)
#define PT_USER_MASK (1ULL << PT_USER_SHIFT)

#define PT_ACCESSED_SHIFT 5
#define PT_ACCESSED_MASK (1ULL << PT_ACCESSED_SHIFT)
#define PT_DIRTY_SHIFT 6
#define PT_DIRTY_MASK (1ULL << PT_DIRTY_SHIFT)
#define PT_PAGE_SIZE_SHIFT 7
#define PT_PAGE_SIZE_MASK (1ULL << PT_PAGE_SIZE_SHIFT)

#define PT64_NX_SHIFT 63

#define PT32_DIR_PSE36_SIZE 4
#define PT32_DIR_PSE36_SHIFT 13
#define PT32_DIR_PSE36_MASK \
	(((1ULL << PT32_DIR_PSE36_SIZE) - 1) << PT32_DIR_PSE36_SHIFT)

#define PT64_ROOT_LEVEL 4
#define PT32_ROOT_LEVEL 2
#define PT32E_ROOT_LEVEL 3

#define PT_DIRECTORY_LEVEL 2
#define PT_PAGE_TABLE_LEVEL 1

static inline unsigned long kvm_mmu_available_pages(struct kvm *kvm)
{
	if (kvm->arch.n_max_mmu_pages > kvm->arch.n_used_mmu_pages)
		return kvm->arch.n_max_mmu_pages -
			kvm->arch.n_used_mmu_pages;

	return 0;
}

static inline int is_present_gpte(unsigned long pte)
{
	return pte & PT_PRESENT_MASK;
}

static inline int is_writable_pte(unsigned long pte)
{
	return pte & PT_WRITABLE_MASK;
}

static inline bool kgrr_is_write_protection(struct kvm_vcpu *vcpu)
{
	return kgrr_kvm_read_cr0_bits(vcpu, X86_CR0_WP);
}

static inline bool kgrr_permission_fault(struct kvm_vcpu *vcpu, struct kvm_mmu *mmu,
				    unsigned pte_access, unsigned pte_pkey,
				    unsigned pfec)
{
	int cpl = (*kgre_kvm_x86_ops)->get_cpl(vcpu);
	unsigned long rflags = (*kgre_kvm_x86_ops)->get_rflags(vcpu);

	/*
	 * If CPL < 3, SMAP prevention are disabled if EFLAGS.AC = 1.
	 *
	 * If CPL = 3, SMAP applies to all supervisor-mode data accesses
	 * (these are implicit supervisor accesses) regardless of the value
	 * of EFLAGS.AC.
	 *
	 * This computes (cpl < 3) && (rflags & X86_EFLAGS_AC), leaving
	 * the result in X86_EFLAGS_AC. We then insert it in place of
	 * the PFERR_RSVD_MASK bit; this bit will always be zero in pfec,
	 * but it will be one in index if SMAP checks are being overridden.
	 * It is important to keep this branchless.
	 */
	unsigned long smap = (cpl - 3) & (rflags & X86_EFLAGS_AC);
	int index = (pfec >> 1) +
		    (smap >> (X86_EFLAGS_AC_BIT - PFERR_RSVD_BIT + 1));
	bool fault = (mmu->permissions[index] >> pte_access) & 1;

	WARN_ON(pfec & (PFERR_PK_MASK | PFERR_RSVD_MASK));

	if (unlikely(mmu->pkru_mask)) {
		u32 pkru_bits, offset;

		/*
		* PKRU defines 32 bits, there are 16 domains and 2
		* attribute bits per domain in pkru.  pte_pkey is the
		* index of the protection domain, so pte_pkey * 2 is
		* is the index of the first bit for the domain.
		*/
		pkru_bits = (kgrr_kvm_read_pkru(vcpu) >> (pte_pkey * 2)) & 3;

		/* clear present bit, replace PFEC.RSVD with ACC_USER_MASK. */
		offset = pfec - 1 +
			((pte_access & PT_USER_MASK) << (PFERR_RSVD_BIT - PT_USER_SHIFT));

		pkru_bits &= mmu->pkru_mask >> offset;
		pfec |= -pkru_bits & PFERR_PK_MASK;
		fault |= (pkru_bits != 0);
	}

	return fault;
}

/* klp-ccp: from arch/x86/kvm/x86.h */
#include <asm/processor.h>
#include <linux/kvm_host.h>

static inline bool mmu_is_nested(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.walk_mmu == &vcpu->arch.nested_mmu;
}

/* klp-ccp: from arch/x86/kvm/mmu.c */
#include <linux/kvm_host.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/compiler.h>
#include <linux/srcu.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <asm/page.h>
#include <asm/cmpxchg.h>
#include <asm/io.h>
#include <asm/vmx.h>

static int __read_mostly (*kgre_nx_huge_pages);

enum {
	AUDIT_PRE_PAGE_FAULT,
	AUDIT_POST_PAGE_FAULT,
	AUDIT_PRE_PTE_WRITE,
	AUDIT_POST_PTE_WRITE,
	AUDIT_PRE_SYNC,
	AUDIT_POST_SYNC
};

#define pgprintk(x...) do { } while (0)

#define PTE_PREFETCH_NUM		8

#define PT64_LEVEL_BITS 9

#define PT64_LEVEL_SHIFT(level) \
		(PAGE_SHIFT + (level - 1) * PT64_LEVEL_BITS)

#define PT64_INDEX(address, level)\
	(((address) >> PT64_LEVEL_SHIFT(level)) & ((1 << PT64_LEVEL_BITS) - 1))

#define PT32_LEVEL_BITS 10

#define PT32_LEVEL_SHIFT(level) \
		(PAGE_SHIFT + (level - 1) * PT32_LEVEL_BITS)

#define PT32_LVL_OFFSET_MASK(level) \
	(PT32_BASE_ADDR_MASK & ((1ULL << (PAGE_SHIFT + (((level) - 1) \
						* PT32_LEVEL_BITS))) - 1))

#define PT32_INDEX(address, level)\
	(((address) >> PT32_LEVEL_SHIFT(level)) & ((1 << PT32_LEVEL_BITS) - 1))

#define PT64_BASE_ADDR_MASK (((1ULL << 52) - 1) & ~(u64)(PAGE_SIZE-1))

#define PT64_LVL_ADDR_MASK(level) \
	(PT64_BASE_ADDR_MASK & ~((1ULL << (PAGE_SHIFT + (((level) - 1) \
						* PT64_LEVEL_BITS))) - 1))
#define PT64_LVL_OFFSET_MASK(level) \
	(PT64_BASE_ADDR_MASK & ((1ULL << (PAGE_SHIFT + (((level) - 1) \
						* PT64_LEVEL_BITS))) - 1))

#define PT32_BASE_ADDR_MASK PAGE_MASK

#define PT32_LVL_ADDR_MASK(level) \
	(PAGE_MASK & ~((1ULL << (PAGE_SHIFT + (((level) - 1) \
					    * PT32_LEVEL_BITS))) - 1))

#define ACC_EXEC_MASK    1
#define ACC_WRITE_MASK   PT_WRITABLE_MASK
#define ACC_USER_MASK    PT_USER_MASK
#define ACC_ALL          (ACC_EXEC_MASK | ACC_WRITE_MASK | ACC_USER_MASK)

#define SHADOW_PT_INDEX(addr, level) PT64_INDEX(addr, level)

#define PTE_LIST_EXT 3

enum {
	RET_PF_RETRY = 0,
	RET_PF_EMULATE = 1,
	RET_PF_INVALID = 2,
};

struct pte_list_desc {
	u64 *sptes[PTE_LIST_EXT];
	struct pte_list_desc *more;
};

struct kvm_shadow_walk_iterator {
	u64 addr;
	hpa_t shadow_addr;
	u64 *sptep;
	int level;
	unsigned index;
};

static u64 __read_mostly (*kgre_shadow_mmio_mask);

/* klp-ccp: from arch/x86/kvm/mmutrace.h */
KGRR_TRACE_EVENT(
	kvm_mmu_pagetable_walk,
	TP_PROTO(u64 addr, u32 pferr),
	TP_ARGS(addr, pferr)
);

KGRR_TRACE_EVENT(
	kvm_mmu_paging_element,
	TP_PROTO(u64 pte, int level),
	TP_ARGS(pte, level)
);


static void (*kgre_trace_kvm_mmu_set_accessed_bit)(unsigned long table_gfn, unsigned index, unsigned size);

static void (*kgre_trace_kvm_mmu_set_dirty_bit)(unsigned long table_gfn, unsigned index, unsigned size);

KGRR_TRACE_EVENT(
	kvm_mmu_walker_error,
	TP_PROTO(u32 pferr),
	TP_ARGS(pferr)
);

KGRR_TRACE_EVENT(
	kvm_mmu_spte_requested,
	TP_PROTO(gpa_t addr, int level, kvm_pfn_t pfn),
	TP_ARGS(addr, level, pfn)
);

/* klp-ccp: from arch/x86/kvm/mmu.c */
static bool kgrr_is_nx_huge_page_enabled(void)
{
	return READ_ONCE((*kgre_nx_huge_pages));
}

static bool kgrr_is_mmio_spte(u64 spte)
{
	return (spte & (*kgre_shadow_mmio_mask)) == (*kgre_shadow_mmio_mask);
}

static int is_cpuid_PSE36(void)
{
	return 1;
}

static int kgrr_is_shadow_present_pte(u64 pte)
{
	return pte & PT_PRESENT_MASK && !kgrr_is_mmio_spte(pte);
}

static int is_large_pte(u64 pte)
{
	return pte & PT_PAGE_SIZE_MASK;
}

static int kgrr_is_rmap_spte(u64 pte)
{
	return kgrr_is_shadow_present_pte(pte);
}

static int is_last_spte(u64 pte, int level)
{
	if (level == PT_PAGE_TABLE_LEVEL)
		return 1;
	if (is_large_pte(pte))
		return 1;
	return 0;
}

static gfn_t pse36_gfn_delta(u32 gpte)
{
	int shift = 32 - PT32_DIR_PSE36_SHIFT - PAGE_SHIFT;

	return (gpte & PT32_DIR_PSE36_MASK) << shift;
}

#ifdef CONFIG_X86_64

static void __update_clear_spte_fast(u64 *sptep, u64 spte)
{
	*sptep = spte;
}

#else
#error "klp-ccp: non-taken branch"
#endif

static void mmu_spte_clear_no_track(u64 *sptep)
{
	__update_clear_spte_fast(sptep, 0ull);
}

static int (*kgre_mmu_topup_memory_caches)(struct kvm_vcpu *vcpu);

static void account_huge_nx_page(struct kvm *kvm, struct kvm_mmu_page *sp)
{
	if (sp->lpage_disallowed)
		return;

	++kvm->stat.nx_lpage_splits;
	list_add_tail(&sp->lpage_disallowed_link,
		      &kvm->arch.lpage_disallowed_mmu_pages);
	sp->lpage_disallowed = true;
}

static int (*kgre_mapping_level)(struct kvm_vcpu *vcpu, gfn_t large_gfn,
			 bool *force_pt_level);

static void (*kgre_pte_list_remove)(u64 *spte, struct kvm_rmap_head *rmap_head);

static struct kvm_rmap_head *__gfn_to_rmap(gfn_t gfn, int level,
					   struct kvm_memory_slot *slot)
{
	unsigned long idx;

	idx = gfn_to_index(gfn, slot->base_gfn, level);
	return &slot->arch.rmap[level - PT_PAGE_TABLE_LEVEL][idx];
}

struct rmap_iterator {
	/* private fields */
	struct pte_list_desc *desc;	/* holds the sptep if not NULL */
	int pos;			/* index of the sptep */
};

static u64 *rmap_get_first(struct kvm_rmap_head *rmap_head,
			   struct rmap_iterator *iter)
{
	if (!rmap_head->val)
		return NULL;

	if (!(rmap_head->val & 1)) {
		iter->desc = NULL;
		return (u64 *)rmap_head->val;
	}

	iter->desc = (struct pte_list_desc *)(rmap_head->val & ~1ul);
	iter->pos = 0;
	return iter->desc->sptes[iter->pos];
}

static u64 *(*kgre_rmap_get_next)(struct rmap_iterator *iter);

#define kgrr_for_each_rmap_spte(_rmap_head_, _iter_, _spte_)			\
	for (_spte_ = rmap_get_first(_rmap_head_, _iter_);		\
	     _spte_ && ({BUG_ON(!kgrr_is_shadow_present_pte(*_spte_)); 1;});	\
	     _spte_ = (*kgre_rmap_get_next)(_iter_))

static void (*kgre_drop_spte)(struct kvm *kvm, u64 *sptep);

static bool kgrr___drop_large_spte(struct kvm *kvm, u64 *sptep)
{
	if (is_large_pte(*sptep)) {
		WARN_ON(page_header(__pa(sptep))->role.level ==
			PT_PAGE_TABLE_LEVEL);
		(*kgre_drop_spte)(kvm, sptep);
		--kvm->stat.lpages;
		return true;
	}

	return false;
}

static void kgrr_drop_large_spte(struct kvm_vcpu *vcpu, u64 *sptep)
{
	if (kgrr___drop_large_spte(vcpu->kvm, sptep))
		(*kgre_kvm_flush_remote_tlbs)(vcpu->kvm);
}

static void kgrr_mmu_page_remove_parent_pte(struct kvm_mmu_page *sp,
				       u64 *parent_pte)
{
	(*kgre_pte_list_remove)(parent_pte, &sp->parent_ptes);
}

static void kgrr_drop_parent_pte(struct kvm_mmu_page *sp,
			    u64 *parent_pte)
{
	kgrr_mmu_page_remove_parent_pte(sp, parent_pte);
	mmu_spte_clear_no_track(parent_pte);
}

static int (*kgre_kvm_mmu_prepare_zap_page)(struct kvm *kvm, struct kvm_mmu_page *sp,
				    struct list_head *invalid_list);
static void (*kgre_kvm_mmu_commit_zap_page)(struct kvm *kvm,
				    struct list_head *invalid_list);

#ifdef CONFIG_KVM_MMU_AUDIT

/* klp-ccp: from arch/x86/kvm/mmu_audit.c */
#include <linux/ratelimit.h>

static char const *(*kgre_audit_point_name)[];

#define kgrr_audit_printk(kvm, fmt, args...)		\
	printk(KERN_ERR "audit: (%s) error: "	\
		fmt, (*kgre_audit_point_name)[kvm->arch.audit_point], ##args)

typedef void (*inspect_spte_fn) (struct kvm_vcpu *vcpu, u64 *sptep, int level);

static void kgrr___mmu_spte_walk(struct kvm_vcpu *vcpu, struct kvm_mmu_page *sp,
			    inspect_spte_fn fn, int level)
{
	int i;

	for (i = 0; i < PT64_ENT_PER_PAGE; ++i) {
		u64 *ent = sp->spt;

		fn(vcpu, ent + i, level);

		if (kgrr_is_shadow_present_pte(ent[i]) &&
		      !is_last_spte(ent[i], level)) {
			struct kvm_mmu_page *child;

			child = page_header(ent[i] & PT64_BASE_ADDR_MASK);
			kgrr___mmu_spte_walk(vcpu, child, fn, level - 1);
		}
	}
}

static void kgrr_mmu_spte_walk(struct kvm_vcpu *vcpu, inspect_spte_fn fn)
{
	int i;
	struct kvm_mmu_page *sp;

	if (!VALID_PAGE(vcpu->arch.mmu.root_hpa))
		return;

	if (vcpu->arch.mmu.root_level == PT64_ROOT_LEVEL) {
		hpa_t root = vcpu->arch.mmu.root_hpa;

		sp = page_header(root);
		kgrr___mmu_spte_walk(vcpu, sp, fn, PT64_ROOT_LEVEL);
		return;
	}

	for (i = 0; i < 4; ++i) {
		hpa_t root = vcpu->arch.mmu.pae_root[i];

		if (root && VALID_PAGE(root)) {
			root &= PT64_BASE_ADDR_MASK;
			sp = page_header(root);
			kgrr___mmu_spte_walk(vcpu, sp, fn, 2);
		}
	}

	return;
}

typedef void (*sp_handler) (struct kvm *kvm, struct kvm_mmu_page *sp);

static void walk_all_active_sps(struct kvm *kvm, sp_handler fn)
{
	struct kvm_mmu_page *sp;

	list_for_each_entry(sp, &kvm->arch.active_mmu_pages, link)
		fn(kvm, sp);
}

static void (*kgre_inspect_spte_has_rmap)(struct kvm *kvm, u64 *sptep);

static void kgrr_check_mappings_rmap(struct kvm *kvm, struct kvm_mmu_page *sp)
{
	int i;

	if (sp->role.level != PT_PAGE_TABLE_LEVEL)
		return;

	for (i = 0; i < PT64_ENT_PER_PAGE; ++i) {
		if (!kgrr_is_rmap_spte(sp->spt[i]))
			continue;

		(*kgre_inspect_spte_has_rmap)(kvm, sp->spt + i);
	}
}

static void kgrr_audit_write_protection(struct kvm *kvm, struct kvm_mmu_page *sp)
{
	struct kvm_rmap_head *rmap_head;
	u64 *sptep;
	struct rmap_iterator iter;
	struct kvm_memslots *slots;
	struct kvm_memory_slot *slot;

	if (sp->role.direct || sp->unsync || sp->role.invalid)
		return;

	slots = kvm_memslots_for_spte_role(kvm, sp->role);
	slot = __gfn_to_memslot(slots, sp->gfn);
	rmap_head = __gfn_to_rmap(sp->gfn, PT_PAGE_TABLE_LEVEL, slot);

	kgrr_for_each_rmap_spte(rmap_head, &iter, sptep) {
		if (is_writable_pte(*sptep))
			kgrr_audit_printk(kvm, "shadow page has writable "
				     "mappings: gfn %llx role %x\n",
				     sp->gfn, sp->role.word);
	}
}

static void kgrr_audit_sp(struct kvm *kvm, struct kvm_mmu_page *sp)
{
	kgrr_check_mappings_rmap(kvm, sp);
	kgrr_audit_write_protection(kvm, sp);
}

static void kgrr_audit_all_active_sps(struct kvm *kvm)
{
	walk_all_active_sps(kvm, kgrr_audit_sp);
}

static void (*kgre_audit_spte)(struct kvm_vcpu *vcpu, u64 *sptep, int level);

static void kgrr_audit_vcpu_spte(struct kvm_vcpu *vcpu)
{
	kgrr_mmu_spte_walk(vcpu, (*kgre_audit_spte));
}

static struct static_key (*kgre_mmu_audit_key);

static void kgrr___kvm_mmu_audit(struct kvm_vcpu *vcpu, int point)
{
	static DEFINE_RATELIMIT_STATE(ratelimit_state, 5 * HZ, 10);

	if (!__ratelimit(&ratelimit_state))
		return;

	vcpu->kvm->arch.audit_point = point;
	kgrr_audit_all_active_sps(vcpu->kvm);
	kgrr_audit_vcpu_spte(vcpu);
}

static inline void kgrr_kvm_mmu_audit(struct kvm_vcpu *vcpu, int point)
{
	if (static_key_enabled((&(*kgre_mmu_audit_key))))
		kgrr___kvm_mmu_audit(vcpu, point);
}

#else
#error "klp-ccp: non-taken branch"
#endif

static void __clear_sp_write_flooding_count(struct kvm_mmu_page *sp)
{
	sp->write_flooding_count = 0;
}

static void clear_sp_write_flooding_count(u64 *spte)
{
	struct kvm_mmu_page *sp =  page_header(__pa(spte));

	__clear_sp_write_flooding_count(sp);
}

static struct kvm_mmu_page *(*kgre_kvm_mmu_get_page)(struct kvm_vcpu *vcpu,
					     gfn_t gfn,
					     gva_t gaddr,
					     unsigned level,
					     int direct,
					     unsigned access);

static void (*kgre_shadow_walk_init)(struct kvm_shadow_walk_iterator *iterator,
			     struct kvm_vcpu *vcpu, u64 addr);

static bool shadow_walk_okay(struct kvm_shadow_walk_iterator *iterator)
{
	if (iterator->level < PT_PAGE_TABLE_LEVEL)
		return false;

	iterator->index = SHADOW_PT_INDEX(iterator->addr, iterator->level);
	iterator->sptep	= ((u64 *)__va(iterator->shadow_addr)) + iterator->index;
	return true;
}

static void __shadow_walk_next(struct kvm_shadow_walk_iterator *iterator,
			       u64 spte)
{
	if (is_last_spte(spte, iterator->level)) {
		iterator->level = 0;
		return;
	}

	iterator->shadow_addr = spte & PT64_BASE_ADDR_MASK;
	--iterator->level;
}

static void shadow_walk_next(struct kvm_shadow_walk_iterator *iterator)
{
	return __shadow_walk_next(iterator, *iterator->sptep);
}

static void (*kgre_link_shadow_page)(struct kvm_vcpu *vcpu, u64 *sptep,
			     struct kvm_mmu_page *sp);

static void kgrr_validate_direct_spte(struct kvm_vcpu *vcpu, u64 *sptep,
				   unsigned direct_access)
{
	if (kgrr_is_shadow_present_pte(*sptep) && !is_large_pte(*sptep)) {
		struct kvm_mmu_page *child;

		/*
		 * For the direct sp, if the guest pte's dirty bit
		 * changed form clean to dirty, it will corrupt the
		 * sp's access: allow writable in the read-only sp,
		 * so we should update the spte at this point to get
		 * a new sp with the correct access.
		 */
		child = page_header(*sptep & PT64_BASE_ADDR_MASK);
		if (child->role.access == direct_access)
			return;

		kgrr_drop_parent_pte(child, sptep);
		(*kgre_kvm_flush_remote_tlbs)(vcpu->kvm);
	}
}

static int (*kgre_kvm_mmu_prepare_zap_page)(struct kvm *kvm, struct kvm_mmu_page *sp,
				    struct list_head *invalid_list);

static void (*kgre_kvm_mmu_commit_zap_page)(struct kvm *kvm,
				    struct list_head *invalid_list);

static bool kgrr_prepare_zap_oldest_mmu_page(struct kvm *kvm,
					struct list_head *invalid_list)
{
	struct kvm_mmu_page *sp;

	if (list_empty(&kvm->arch.active_mmu_pages))
		return false;

	sp = list_entry(kvm->arch.active_mmu_pages.prev,
			struct kvm_mmu_page, link);
	(*kgre_kvm_mmu_prepare_zap_page)(kvm, sp, invalid_list);

	return true;
}

static int (*kgre_mmu_set_spte)(struct kvm_vcpu *vcpu, u64 *sptep, unsigned pte_access,
			int write_fault, int level, gfn_t gfn, kvm_pfn_t pfn,
		       	bool speculative, bool host_writable);

static void (*kgre___direct_pte_prefetch)(struct kvm_vcpu *vcpu,
				  struct kvm_mmu_page *sp, u64 *sptep);

static void kgrr_disallowed_hugepage_adjust(struct kvm_shadow_walk_iterator it,
				       gfn_t gfn, kvm_pfn_t *pfnp, int *levelp)
{
	int level = *levelp;
	u64 spte = *it.sptep;

	if (it.level == level && level > PT_PAGE_TABLE_LEVEL &&
	    kgrr_is_nx_huge_page_enabled() &&
	    kgrr_is_shadow_present_pte(spte) &&
	    !is_large_pte(spte)) {
		/*
		 * A small SPTE exists for this pfn, but FNAME(fetch)
		 * and __direct_map would like to create a large PTE
		 * instead: just force them to go down another level,
		 * patching back for them into pfn the next 9 bits of
		 * the address.
		 */
		u64 page_mask = KVM_PAGES_PER_HPAGE(level) - KVM_PAGES_PER_HPAGE(level - 1);
		*pfnp |= gfn & page_mask;
		(*levelp)--;
	}
}

static void (*kgre_transparent_hugepage_adjust)(struct kvm_vcpu *vcpu,
					gfn_t gfn, kvm_pfn_t *pfnp,
					int *levelp);

static bool (*kgre_handle_abnormal_pfn)(struct kvm_vcpu *vcpu, gva_t gva, gfn_t gfn,
				kvm_pfn_t pfn, unsigned access, int *ret_val);

static bool (*kgre_try_async_pf)(struct kvm_vcpu *vcpu, bool prefault, gfn_t gfn,
			 gva_t gva, kvm_pfn_t *pfn, bool write, bool *writable);
static void kgrr_make_mmu_pages_available(struct kvm_vcpu *vcpu);

static bool
__is_rsvd_bits_set(struct rsvd_bits_validate *rsvd_check, u64 pte, int level)
{
	int bit7 = (pte >> 7) & 1, low6 = pte & 0x3f;

	return (pte & rsvd_check->rsvd_bits_mask[bit7][level-1]) |
		((rsvd_check->bad_mt_xwr & (1ull << low6)) != 0);
}

static bool is_rsvd_bits_set(struct kvm_mmu *mmu, u64 gpte, int level)
{
	return __is_rsvd_bits_set(&mmu->guest_rsvd_check, gpte, level);
}

static bool (*kgre_try_async_pf)(struct kvm_vcpu *vcpu, bool prefault, gfn_t gfn,
			 gva_t gva, kvm_pfn_t *pfn, bool write, bool *writable);

static void inject_page_fault(struct kvm_vcpu *vcpu,
			      struct x86_exception *fault)
{
	vcpu->arch.mmu.inject_page_fault(vcpu, fault);
}

static inline bool is_last_gpte(struct kvm_mmu *mmu,
				unsigned level, unsigned gpte)
{
	/*
	 * The RHS has bit 7 set iff level < mmu->last_nonleaf_level.
	 * If it is clear, there are no large pages at this level, so clear
	 * PT_PAGE_SIZE_MASK in gpte if that is the case.
	 */
	gpte &= level - mmu->last_nonleaf_level;

	/*
	 * PT_PAGE_TABLE_LEVEL always terminates.  The RHS has bit 7 set
	 * iff level <= PT_PAGE_TABLE_LEVEL, which for our purpose means
	 * level == PT_PAGE_TABLE_LEVEL; set PT_PAGE_SIZE_MASK in gpte then.
	 */
	gpte |= level - PT_PAGE_TABLE_LEVEL - 1;

	return gpte & PT_PAGE_SIZE_MASK;
}

#define PTTYPE_EPT 18 /* arbitrary */
#define PTTYPE PTTYPE_EPT
#include "bsc1189278_paging_tmpl.h"
#undef PTTYPE

#define PTTYPE 64
#include "bsc1189278_paging_tmpl.h"
#undef PTTYPE

#define PTTYPE 32
#include "bsc1189278_paging_tmpl.h"
#undef PTTYPE

static void kgrr_make_mmu_pages_available(struct kvm_vcpu *vcpu)
{
	LIST_HEAD(invalid_list);

	if (likely(kvm_mmu_available_pages(vcpu->kvm) >= KVM_MIN_FREE_MMU_PAGES))
		return;

	while (kvm_mmu_available_pages(vcpu->kvm) < KVM_REFILL_PAGES) {
		if (!kgrr_prepare_zap_oldest_mmu_page(vcpu->kvm, &invalid_list))
			break;

		++vcpu->kvm->stat.mmu_recycled;
	}
	(*kgre_kvm_mmu_commit_zap_page)(vcpu->kvm, &invalid_list);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1189278.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "kvm"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "nx_huge_pages", (void *)&kgre_nx_huge_pages, "kvm" },
	{ "__tracepoint_kvm_mmu_paging_element",
	  (void *)&kgre___tracepoint_kvm_mmu_paging_element, "kvm" },
	{ "shadow_mmio_mask", (void *)&kgre_shadow_mmio_mask, "kvm" },
	{ "__tracepoint_kvm_mmu_pagetable_walk",
	  (void *)&kgre___tracepoint_kvm_mmu_pagetable_walk, "kvm" },
	{ "__tracepoint_kvm_mmu_walker_error",
	  (void *)&kgre___tracepoint_kvm_mmu_walker_error, "kvm" },
	{ "__tracepoint_kvm_mmu_spte_requested",
	  (void *)&kgre___tracepoint_kvm_mmu_spte_requested, "kvm" },
	{ "audit_point_name", (void *)&kgre_audit_point_name, "kvm" },
	{ "mmu_audit_key", (void *)&kgre_mmu_audit_key, "kvm" },
	{ "kvm_x86_ops", (void *)&kgre_kvm_x86_ops, "kvm" },
	{ "trace_kvm_mmu_set_accessed_bit",
	  (void *)&kgre_trace_kvm_mmu_set_accessed_bit, "kvm" },
	{ "trace_kvm_mmu_set_dirty_bit",
	  (void *)&kgre_trace_kvm_mmu_set_dirty_bit, "kvm" },
	{ "mmu_topup_memory_caches", (void *)&kgre_mmu_topup_memory_caches,
	  "kvm" },
	{ "mapping_level", (void *)&kgre_mapping_level, "kvm" },
	{ "pte_list_remove", (void *)&kgre_pte_list_remove, "kvm" },
	{ "rmap_get_next", (void *)&kgre_rmap_get_next, "kvm" },
	{ "drop_spte", (void *)&kgre_drop_spte, "kvm" },
	{ "kvm_mmu_prepare_zap_page", (void *)&kgre_kvm_mmu_prepare_zap_page,
	  "kvm" },
	{ "kvm_mmu_commit_zap_page", (void *)&kgre_kvm_mmu_commit_zap_page,
	  "kvm" },
	{ "inspect_spte_has_rmap", (void *)&kgre_inspect_spte_has_rmap, "kvm" },
	{ "audit_spte", (void *)&kgre_audit_spte, "kvm" },
	{ "link_shadow_page", (void *)&kgre_link_shadow_page, "kvm" },
	{ "kvm_mmu_get_page", (void *)&kgre_kvm_mmu_get_page, "kvm" },
	{ "shadow_walk_init", (void *)&kgre_shadow_walk_init, "kvm" },
	{ "mmu_set_spte", (void *)&kgre_mmu_set_spte, "kvm" },
	{ "__direct_pte_prefetch", (void *)&kgre___direct_pte_prefetch, "kvm" },
	{ "handle_abnormal_pfn", (void *)&kgre_handle_abnormal_pfn, "kvm" },
	{ "transparent_hugepage_adjust",
	  (void *)&kgre_transparent_hugepage_adjust, "kvm" },
	{ "try_async_pf", (void *)&kgre_try_async_pf, "kvm" },
	{ "ept_prefetch_gpte", (void *)&kgre_ept_prefetch_gpte, "kvm" },
	{ "paging64_prefetch_gpte", (void *)&kgre_paging64_prefetch_gpte,
	  "kvm" },
	{ "paging32_prefetch_gpte", (void *)&kgre_paging32_prefetch_gpte,
	  "kvm" },
	{ "kvm_release_page_dirty", (void *)&kgre_kvm_release_page_dirty,
	  "kvm" },
	{ "kvm_release_pfn_clean", (void *)&kgre_kvm_release_pfn_clean, "kvm" },
	{ "kvm_vcpu_gfn_to_hva_prot", (void *)&kgre_kvm_vcpu_gfn_to_hva_prot,
	  "kvm" },
	{ "kvm_vcpu_read_guest_atomic",
	  (void *)&kgre_kvm_vcpu_read_guest_atomic, "kvm" },
	{ "kvm_vcpu_mark_page_dirty", (void *)&kgre_kvm_vcpu_mark_page_dirty,
	  "kvm" },
	{ "kvm_flush_remote_tlbs", (void *)&kgre_kvm_flush_remote_tlbs, "kvm" },
};

static int kgr_patch_bsc1189278_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1189278_module_nb = {
	.notifier_call = kgr_patch_bsc1189278_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1189278_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1189278_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1189278_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1189278_module_nb);
}

#endif /* IS_ENABLED(CONFIG_X86_64) */
